package cz.cvut.kbss.reporting.service.repository;

import cz.cvut.kbss.reporting.environment.Environment;
import cz.cvut.kbss.reporting.environment.Generator;
import cz.cvut.kbss.reporting.exception.AuthorizationException;
import cz.cvut.kbss.reporting.exception.UsernameExistsException;
import cz.cvut.kbss.reporting.model.Person;
import cz.cvut.kbss.reporting.service.BaseServiceTestRunner;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.springframework.beans.factory.annotation.Autowired;

import java.security.Principal;

import static org.hamcrest.CoreMatchers.containsString;
import static org.junit.Assert.*;

public class PersonServiceTest extends BaseServiceTestRunner {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    @Autowired
    private PersonService personService;

    @Test
    public void persistEncodesPersonsPassword() {
        final Person p = Generator.generatePerson();
        final String rawPassword = p.getPassword();
        personService.persist(p);
        assertNotEquals(rawPassword, p.getPassword());
        assertTrue(passwordEncoder.matches(rawPassword, p.getPassword()));
    }

    @Test
    public void tryingToUpdateDifferentThanCurrentUserThrowsException() {
        final Person current = Generator.generatePerson();
        persistPerson(current);
        Environment.setCurrentUser(current);
        final Person update = Generator.generatePerson();
        update.setId(Generator.randomInt());
        thrown.expect(AuthorizationException.class);
        thrown.expectMessage(containsString("Modifying other user\'s account is forbidden."));
        personService.update(update);
    }

    @Test
    public void updateEncodesPersonPassword() {
        final Person current = Generator.generatePerson();
        persistPerson(current);
        Environment.setCurrentUser(current);
        final Person update = clonePerson(current);
        final String updatedPassword = "updatedPassword";
        update.setPassword(updatedPassword);
        personService.update(update);

        final Person result = personService.find(current.getId());
        assertTrue(passwordEncoder.matches(updatedPassword, result.getPassword()));
    }

    private Person clonePerson(Person orig) {
        final Person copy = new Person();
        copy.setId(orig.getId());
        copy.setFirstName(orig.getFirstName());
        copy.setLastName(orig.getLastName());
        copy.setUsername(orig.getUsername());
        copy.setPassword(orig.getPassword());
        return copy;
    }

    @Test
    public void updateUsesOriginalPasswordWhenUpdateDoesNotContainPassword() {
        final Person current = Generator.generatePerson();
        persistPerson(current);
        Environment.setCurrentUser(current);
        // We have to store the password, because since the whole test is in one transaction, the update changes
        // the state of the current instance as well (it happens in the same persistence context)
        final String originalPassword = current.getPassword();
        final Person update = clonePerson(current);
        final String newName = "UpdatedFirstName";
        update.setFirstName(newName);
        update.setPassword(null);
        personService.update(update);

        final Person result = personService.find(current.getId());
        assertEquals(originalPassword, result.getPassword());
        assertEquals(newName, result.getFirstName());
    }

    @Test
    public void updateReloadsCurrentSecurityContext() {
        final Person current = Generator.generatePerson();
        persistPerson(current);
        Environment.setCurrentUser(current);
        final Principal originalPrincipal = Environment.getCurrentUserPrincipal();
        final Person update = clonePerson(current);
        final String updatedFirstName = "updatedFirstName";
        final String updatedPassword = "updatedPassword";
        update.setFirstName(updatedFirstName);
        update.setPassword(updatedPassword);
        personService.update(update);

        final Person newCurrent = Environment.getCurrentUser();
        assertNotEquals(originalPrincipal, Environment.getCurrentUserPrincipal());
        assertEquals(update, newCurrent);
    }

    @Test
    public void updateFailsWhenUsernameAlreadyExists() {
        final Person current = Generator.generatePerson();
        final Person anotherPerson = Generator.generatePerson();
        persistPerson(anotherPerson);
        persistPerson(current);
        Environment.setCurrentUser(current);
        final Person update = clonePerson(current);
        update.setUsername(anotherPerson.getUsername());
        thrown.expect(UsernameExistsException.class);
        thrown.expectMessage(containsString(update.getUsername() + " already exists"));

        personService.update(update);
    }

    @Test
    public void existsReturnsTrueForExistingUsername() {
        final Person person = Generator.generatePerson();
        persistPerson(person);
        assertTrue(personService.exists(person.getUsername()));
    }
}